from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Dict

import torch

from ...configs import DatasetConfig
from ...constants import Backends, TaskType
from ...registry import register_dataset
from ...utils import Logger, is_backend_available, is_text_valid, reverse_string_digits
from ..data_collators import CharLevelOCRDataCollator
from .dataset import Dataset


if is_backend_available(Backends.DATASETS):
    from datasets import load_dataset

logger = Logger(__name__)

_required_backends = [Backends.SCIKIT]

fa_characters = [
    "", "آ", "ا", "ب", "پ", "ت", "ث", "ج", "چ", "ح", "خ", "د", "ذ", "ر", "ز", "ژ", "س", "ش",
    "ص", "ض", "ط", "ظ", "ع", "غ", "ف", "ق", "ک", "گ", "ل", "م", "ن", "و", "ه", "ی", " "
]
fa_numbers = ["۱", "۲", "۳", "۴", "۵", "۶", "۷", "۸", "۹", "۰"]
fa_special_characters = ["ء", "ؤ", "ئ", "أ", "ّ"]
fa_symbols = ["/", "(", ")", "+", "-", ":", "،", "!", ".", "؛", "=", "%", "؟"]
en_numbers = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "0"]
all_characters = fa_characters + fa_numbers + fa_special_characters + fa_symbols + en_numbers

ID2LABEL = dict(enumerate(all_characters))


class TextSplitType(str, Enum):
    CHAR_SPLIT = "char_split"  # mostly for char level ocr models
    TOKENIZE = "tokenize"  # mostly for transformer-based ocr models


@dataclass
class OCRDatasetConfig(DatasetConfig):
    """
    Configuration class for OCR datasets.

    Args:
        path (str): Path to the dataset.
        text_split_type (TextSplitType): Type of text splitting (CHAR_SPLIT or TOKENIZE).
        id2label (Dict[int, str]): Mapping of label IDs to characters.
        text_column (str): Column name for text in the dataset.
        images_paths_column (str): Column name for image paths in the dataset.
        max_length (int): Maximum length of text.
        invalid_characters (list): List of invalid characters.
        reverse_digits (bool): Whether to reverse the digits in text.

    """
    name = "ocr"
    task: TaskType = TaskType.IMAGE2TEXT
    path: str = None
    text_split_type: str | TextSplitType = TextSplitType.CHAR_SPLIT
    id2label: Dict[int, str] = field(default_factory=lambda: ID2LABEL)
    text_column: str = "label"
    images_paths_column: str = "image_path"
    max_length: int = None
    invalid_characters: list = None
    reverse_text: bool = None
    reverse_digits: bool = None


@register_dataset("ocr", config_class=OCRDatasetConfig)
class OCRDataset(Dataset):
    """
    General OCR dataset class.

    OCR dataset supports two types of image to text dataset. One is for tokenizer-based models in which the labels are
    tokens and the other is char-level models in which the labels are separated by character and the converted to ids.
    This behavior is specified by the `text_split_type` in config which can be either `tokenize` or `char_split`.

    """
    required_backends = _required_backends

    def __init__(self, config: OCRDatasetConfig, split=None, preprocessor=None, **kwargs):
        super().__init__(config=config, split=split, preprocessor=preprocessor, **kwargs)
        self.image_processor = self.preprocessor.image_processor
        if self.config.text_split_type == TextSplitType.TOKENIZE:
            if self.config.tokenizer_path is not None:
                self.tokenizer = self.preprocessor.tokenizer
                self.data_collator = None  # TODO resolve this in the future.
            else:
                raise ValueError("No `tokenizer_path` given although `text_split_type` is set to `tokenize`!")
        else:
            self.tokenizer = None
            self.data_collator = CharLevelOCRDataCollator()

    def _load(self, split=None):
        """
        Load the dataset and clean up invalid samples.

        Args:
            split: Dataset split, defaults to None.

        Returns:
            Dataset: The cleaned dataset.

        """
        data = load_dataset(self.config.path, split=split, cache_dir=self.cache_dir, **self.config.hf_load_kwargs)
        # Cleanup dataset
        valid_indices = []
        invalid_indices = []
        for i, sample in enumerate(list(iter(data))):
            path, text = sample.values()
            if len(text) <= self.config.max_length and is_text_valid(text, self.config.id2label.values()):
                valid_indices.append(i)
            else:
                invalid_indices.append(i)
        if len(invalid_indices):
            logger.warning(
                f"{len(invalid_indices)} invalid samples found in the dataset! "
                f"Inspect them using the `invalid_data` attribute"
            )
        self.invalid_data = data.select(invalid_indices)
        data = data.select(valid_indices)
        return data

    def _text_to_tensor(self, text):
        """
        Convert text to tensor based on the configured text_split_type.

        Args:
            text (str): The raw text.

        Returns:
            torch.Tensor: The output tensor.

        """
        # Tokenize text inputs if text_split_type is set to `tokenize`
        if self.config.text_split_type == TextSplitType.TOKENIZE:
            token_ids = self.tokenizer(text, padding="max_length", max_length=self.config.max_length)["token_ids"]
            # Make sure to ignore pad tokens by the loss function
            token_ids = [token_id if token_id != self.tokenizer.pad_token_id else -100 for token_id in token_ids]
            labels = torch.tensor(token_ids)
        # If split text is not tokenizer-based
        elif self.config.text_split_type == TextSplitType.CHAR_SPLIT:
            if self.config.reverse_digits:
                text = reverse_string_digits(text)
            label2id = {v: k for k, v in self.config.id2label.items()}
            labels = [label2id[x] for x in text]
            labels = torch.LongTensor(labels)
        else:
            raise ValueError(f"Invalid `text_split_type={self.config.text_split_type}`")

        return labels

    def __getitem__(self, index):
        """
        Get a specific item from the dataset.

        Args:
            index: Index of the item to retrieve.

        Returns:
            dict: The input data.

        """
        path, text = self.data[index].values()
        pixel_values = self.image_processor(path, return_tensors="torch")["pixel_values"][0]
        labels = self._text_to_tensor(text)
        inputs = {
            "pixel_values": pixel_values,
            "labels": labels,
        }
        return inputs



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterable

from ..configs import MetricConfig
from ..constants import Backends, MetricType
from ..registry import register_metric
from ..utils import is_backend_available
from .metric import Metric


if is_backend_available(Backends.SCIKIT):
    from sklearn.metrics import precision_score

_required_backends = [
    Backends.SCIKIT,
]


@dataclass
class PrecisionConfig(MetricConfig):
    """
    Configuration class for Precision metric.

    Args:
        name (MetricType): The type of metric, Precision in this case.
        pos_label (int): Label of the positive class.
        average (str): Type of averaging for the precision score.
        sample_weight (Iterable[float]): Sample weights for the precision score.
        zero_division (str | float): Strategy for zero-division, default is 0.0.
        output_keys (tuple): Keys to filter the metric results for output.
    """
    name = MetricType.PRECISION
    objective: str = "maximize"
    pos_label: int = 1
    average: str = "macro"
    sample_weight: Iterable[float] = None
    zero_division: str | float = 0.0
    output_keys: tuple = ("precision",)


@register_metric("precision", config_class=PrecisionConfig)
class Precision(Metric):
    """
    Precision metric for evaluating classification performance using sklearn's `precision_score`.

    Args:
        config (PrecisionConfig): Metric configuration object.
        **kwargs: Extra configuration parameters passed as kwargs to update the `config`.
    """
    required_backends = _required_backends

    def __init__(self, config: PrecisionConfig, **kwargs):
        super().__init__(config, **kwargs)

    def compute(
        self,
        predictions=None,
        targets=None,
        labels=None,
        pos_label=1,
        average=None,
        sample_weight=None,
        zero_division=None,
        n_decimals=None,
        output_keys=None,
    ):
        """
        Computes the Precision score for the given predictions against targets.

        Args:
            predictions: Predicted labels.
            targets: Ground truth labels.
            labels: List of labels to include in the calculation.
            pos_label (int): Label of the positive class.
            average (str): Type of averaging for the precision score.
            sample_weight (Iterable[float]): Sample weights for the precision score.
            zero_division (str | float): Strategy for zero-division, default is 0.0.
            n_decimals (int): Number of decimals for the final score.
            output_keys (tuple): Filter the output keys.

        Returns:
            dict: A dictionary of the metric results, with keys specified by `output_keys`.
        """
        pos_label = pos_label or self.config.pos_label
        average = average or self.config.average
        sample_weight = sample_weight or self.config.sample_weight
        zero_division = zero_division or self.config.zero_division
        n_decimals = n_decimals or self.config.n_decimals
        output_keys = output_keys or self.config.output_keys

        score = precision_score(
            targets,
            predictions,
            labels=labels,
            pos_label=pos_label,
            average=average,
            sample_weight=sample_weight,
            zero_division=zero_division,
        )

        score = float(score) if score.size == 1 else score

        results = {"precision": round(float(score), n_decimals)}

        if output_keys:
            results = {k: v for k, v in results.items() if k in output_keys}

        return results



"""
Configs are at the core of Hezar. All core modules like `Model`, `Preprocessor`, `Trainer`, etc. take their parameters
as a config container which is an instance of `Config` or its derivatives. A `Config` is a Python dataclass with
auxiliary methods for loading, saving, uploading to the hub, etc.

Examples:
    >>> from hezar.configs import ModelConfig
    >>> config = ModelConfig.load("hezarai/bert-base-fa")

    >>> from hezar.models import BertMaskFillingConfig
    >>> bert_config = BertMaskFillingConfig(vocab_size=50000, hidden_size=768)
    >>> bert_config.save("saved/bert", filename="model_config.yaml")
    >>> bert_config.push_to_hub("hezarai/bert-custom", filename="model_config.yaml")
"""
from __future__ import annotations

import os
import tempfile
from dataclasses import asdict, dataclass, field, fields
from enum import Enum
from pprint import pformat
from typing import Any, Dict, List, Literal, Optional, Tuple

from huggingface_hub import create_repo, hf_hub_download, upload_file
from omegaconf import DictConfig, OmegaConf

from .constants import (
    DEFAULT_MODEL_CONFIG_FILE,
    HEZAR_CACHE_DIR,
    ConfigType,
    LRSchedulerType,
    OptimizerType,
    PrecisionType,
    TaskType,
)
from .utils import Logger, get_module_config_class


__all__ = [
    "Config",
    "ModelConfig",
    "PreprocessorConfig",
    "TrainerConfig",
    "DatasetConfig",
    "EmbeddingConfig",
    "MetricConfig",
]

logger = Logger(__name__)

CONFIG_CLASS_VARS = ["name", "config_type"]

_config_to_type_mapping = {
    "ModelConfig": ConfigType.MODEL,
    "PreprocessorConfig": ConfigType.PREPROCESSOR,
    "TrainerConfig": ConfigType.TRAINER,
    "DatasetConfig": ConfigType.DATASET,
    "EmbeddingConfig": ConfigType.EMBEDDING,
    "CriterionConfig": ConfigType.CRITERION,
    "MetricConfig": ConfigType.METRIC,
}
_type_to_config_mapping = {v: k for k, v in _config_to_type_mapping.items()}


@dataclass
class Config:
    """
    Base class for all configs in Hezar.

    All configs are simple dataclasses that have some customized functionalities to manage their attributes. There are
    also some Hezar specific methods: `load`, `save` and `push_to_hub`.

    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.BASE)

    def __post_init__(self):
        # Class variables cannot be init-able
        fields_dict = {f.name: f for f in fields(self)}
        for attr in CONFIG_CLASS_VARS:
            if fields_dict[attr].init == True:  # noqa
                raise ValueError(
                    f"The parameter `{attr}` in a config should be either non-initable or unannotated! "
                    f"\nYou should define it as either:\n"
                    f"`{attr} = '{getattr(self, attr)}'`"
                    f" or "
                    f"`{attr}: str = field(default='{getattr(self, attr)}', init=False)`"
                )

        # Convert enums to values
        for param, value in self.dict().items():
            if isinstance(getattr(self, param), Enum):
                setattr(self, param, str(getattr(self, param)))

    def __str__(self):
        return pformat(self.dict())

    def __getitem__(self, item):
        try:
            return self.dict()[item]
        except KeyError:
            raise AttributeError(f"`{self.__class__.__name__}` does not have the parameter `{item}`!")

    def __len__(self):
        return len(self.dict())

    def __iter__(self):
        return iter(self.dict())

    @classmethod
    def fields(cls):
        return cls.__dataclass_fields__

    def dict(self):
        """
        Returns the config object as a dictionary (works on nested dataclasses too)

        Returns:
            The config object as a dictionary
        """
        return asdict(self)

    def keys(self):
        return list(self.dict().keys())

    def get(self, key, default=None):
        return getattr(self, key, default)

    def update(self, d: dict, **kwargs):
        """
        Update config with a given dictionary or keyword arguments. If a key does not exist in the attributes, prints a
        warning but sets it anyway.

        Args:
            d: A dictionary
            **kwargs: Key/value pairs in the form of keyword arguments

        Returns:
            The config object itself but the operation happens in-place anyway
        """
        d.update(kwargs)
        for k, v in d.items():
            if k not in self.fields():
                logger.warning(f"`{str(self.__class__.__name__)}` does not take `{k}` as a config parameter!")
            setattr(self, k, v)
        return self

    @classmethod
    def load(
        cls,
        hub_or_local_path: str | os.PathLike,
        filename: Optional[str] = None,
        subfolder: Optional[str] = None,
        repo_type: str = None,
        cache_dir: str = None,
        **kwargs,
    ) -> "Config":
        """
        Load config from Hub or locally if it already exists on disk (handled by HfApi)

        Args:
            hub_or_local_path: Local or Hub path for the config
            filename: Configuration filename
            subfolder: Optional subfolder path where the config is in
            repo_type: Repo type e.g, model, dataset, etc
            cache_dir: Path to cache directory
            **kwargs: Manual config parameters to override

        Returns:
            A Config instance
        """
        filename = filename or DEFAULT_MODEL_CONFIG_FILE
        subfolder = subfolder or ""

        config_path = os.path.join(hub_or_local_path, subfolder, filename)
        is_local = os.path.isfile(config_path)
        if os.path.isdir(hub_or_local_path) and not is_local:
            raise EnvironmentError(
                f"Path `{hub_or_local_path}` exists locally but the config file {filename} is missing!"
            )
        # if the file or repo_id does not exist locally, load from the Hub
        if not is_local:
            config_path = hf_hub_download(
                hub_or_local_path,
                filename=filename,
                subfolder=subfolder,
                cache_dir=cache_dir or HEZAR_CACHE_DIR,
                repo_type=repo_type,
            )
        # Load config file and convert to dictionary
        dict_config = OmegaConf.load(config_path)
        config = OmegaConf.to_container(dict_config)
        # Check if config_type in the file and class are equal
        config_type = config.get("config_type", ConfigType.BASE)
        if config_type in _config_to_type_mapping.values():
            if config_type != cls.config_type:
                raise ValueError(
                    f"The `config_type` for `{cls.__name__}` is `{cls.config_type}` "
                    f"which is different from the `config_type` parameter in `{filename}` which is `{config_type}`!"
                )
        config_cls = get_module_config_class(config["name"], registry_type=config_type)
        if config_cls is None:
            config_cls = cls
        config = config_cls.from_dict(config, **kwargs)
        return config

    @classmethod
    def from_dict(cls, dict_config: Dict | DictConfig, **kwargs):
        """
        Load config from a dict-like object. Nested configs are also recursively converted to their classes if possible.
        """
        # Update config parameters with kwargs
        dict_config.update(**kwargs)

        for k, v in dict_config.items():
            if isinstance(v, Dict) and "name" in v and "config_type" in v:
                config_cls = get_module_config_class(v["name"], v["config_type"])
                if config_cls is not None:
                    dict_config[k] = config_cls.from_dict(v)

        dict_config = {k: v for k, v in dict_config.items() if k in cls.fields() and cls.fields()[k].init}

        config = cls(**dict_config)  # noqa

        return config

    def save(
        self,
        save_dir: str | os.PathLike,
        filename: str,
        subfolder: Optional[str] = None,
        skip_none_fields: Optional[bool] = True,
    ):
        """
        Save the `*config.yaml` file to a local path

        Args:
             save_dir: Save directory path
             filename: Config file name
             subfolder: Subfolder to save the config file
             skip_none_fields (bool): Whether to skip saving None values or not
        """
        subfolder = subfolder or ""
        config = self.dict()

        if skip_none_fields:
            # exclude None items
            config = {k: v for k, v in config.items() if v is not None}

        # make and save to directory
        os.makedirs(os.path.join(save_dir, subfolder), exist_ok=True)
        save_path = os.path.join(save_dir, subfolder, filename)
        OmegaConf.save(config, save_path)

        return save_path

    def push_to_hub(
        self,
        repo_id: str,
        filename: str,
        subfolder: Optional[str] = None,
        repo_type: Optional[str] = "model",
        skip_none_fields: Optional[bool] = True,
        private: Optional[bool] = False,
        commit_message: Optional[str] = None,
    ):
        """
        Push the config file to the hub

        Args:
            repo_id (str): Repo name or id on the Hub
            filename (str): config file name
            subfolder (str): subfolder to save the config
            repo_type (str): Type of the repo e.g, model, dataset, space
            skip_none_fields (bool): Whether to skip saving None values or not
            private (bool): Whether the repo type should be private or not (ignored if the repo exists)
            commit_message (str): Push commit message
        """
        path_in_repo = f"{subfolder}/{filename}" if subfolder else filename
        subfolder = subfolder or ""

        # create remote repo
        create_repo(repo_id, repo_type=repo_type, private=private, exist_ok=True)
        # save to tmp and prepare for push
        cache_path = tempfile.mkdtemp()
        config_path = self.save(cache_path, filename=filename, subfolder=subfolder, skip_none_fields=skip_none_fields)
        # push to hub
        if commit_message is None:
            commit_message = f"Hezar: Upload {filename}"
        upload_file(
            path_or_fileobj=config_path,
            path_in_repo=path_in_repo,
            repo_id=repo_id,
            repo_type=repo_type,
            commit_message=commit_message,
        )
        logger.log_upload_success(name=f"{self.__class__.__name__}()", target_path=os.path.join(repo_id, path_in_repo))


@dataclass
class ModelConfig(Config):
    """
    Base dataclass for all model configs
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.MODEL)


@dataclass
class PreprocessorConfig(Config):
    """
    Base dataclass for all preprocessor configs
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.PREPROCESSOR)


@dataclass
class DatasetConfig(Config):
    """
    Base dataclass for all dataset configs

    Args:
        path (str):
            Path to the dataset either on the Hub or local. Supported syntax is either `<path>` or `<path>:<name>` where
            <name> is the parameter `name` in the `load_dataset()`
        task (str):
            A supported task for the dataset
        max_size (int | float):
            Maximum number of data samples. Overwrites the main length of the dataset when calling `len(dataset)`.
            If set to a float value between 0 and 1, will be interpreted as a fraction value, e.g, 0.3 means 30% of the
            whole length.
        hf_load_kwargs (dict):
            Keyword arguments to pass to the HF `datasets.load_dataset()`
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.DATASET)
    path: str = None
    task: TaskType | List[TaskType] = field(
        default=None,
        metadata={"help": "Name of the task(s) this dataset is built for"}
    )
    max_size: int | float = None
    hf_load_kwargs: dict = None

    def __post_init__(self):
        super().__post_init__()
        if self.path and ":" in self.path:
            self.path, config_name = self.path.split(":")
            self.hf_load_kwargs["name"] = config_name
        if self.hf_load_kwargs:
            self.hf_load_kwargs.pop("path", None)
            self.hf_load_kwargs.pop("cache_dir", None)
            self.hf_load_kwargs.pop("split", None)
        else:
            self.hf_load_kwargs = {}


@dataclass
class EmbeddingConfig(Config):
    """
    Base dataclass for all embedding configs
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.EMBEDDING)
    bypass_version_check: bool = field(
        default=False,
        metadata={"help": "Whether to bypass checking gensim/numpy/hezar version compatibility"},
    )


@dataclass
class MetricConfig(Config):
    """
    Base dataclass config for all metric configs
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.METRIC)
    objective: Literal["maximize", "minimize"] = None
    output_keys: List | Tuple = None
    n_decimals: int = 4


@dataclass
class TrainerConfig(Config):
    """
    Base dataclass for all trainer configs

    Args:
        task (str, TaskType):
            The training task. Must be a valid name from `TaskType`.
        output_dir (str):
            Path to the directory to save trainer properties.
        device (str):
            Hardware device e.g, `cuda:0`, `cpu`, etc.
        num_epochs (int):
            Number of total epochs to train the model.
        init_weights_from (str):
            Path to a model from disk or Hub to load the initial weights from. Note that this only loads the model
            weights and ignores other checkpoint-related states if the path is a checkpoint. To resume training from
            a checkpoint use the `resume` parameter.
        resume_from_checkpoint (bool, str, os.PathLike):
            Resume training from a checkpoint. If set to True, the trainer will load the latest checkpoint, otherwise if
            a path to a checkpoint is given, it will load that checkpoint and all the other states corresponding to that
            checkpoint.
        max_steps (int):
            Maximum number of iterations to train. This helps to limit how many batches you want to train in total.
        num_dataloader_workers (int):
            Number of dataloader workers, defaults to 4 .
        dataloader_shuffle (bool):
            Control dataloaders `shuffle` argument.
        seed (int):
            Control determinism of the run by setting a seed value. defaults to 42.
        optimizer (OptimizerType):
            Name of the optimizer, available values include properties in `OptimizerType` enum.
        learning_rate (float):
            Initial learning rate for the optimizer.
        weight_decay (float):
            Optimizer weight decay value.
        lr_scheduler (LRSchedulerType):
            Optional learning rate scheduler among `LRSchedulerType` enum.
        lr_scheduler_kwargs (Dict[str, Any]):
            LR scheduler instructor kwargs depending on the scheduler type
        lr_scheduling_steps (int):
            Number of steps to perform scheduler stepping. If left as None, will default to the steps in one full epoch.
        batch_size (int):
            Training batch size.
        eval_batch_size (int):
            Evaluation batch size, defaults to `batch_size` if None.
        gradient_accumulation_steps (int):
            Number of updates steps to accumulate before performing a backward/update pass, defaults to 1.
        distributed (bool):
            Whether to use distributed training (via the `accelerate` package)
        mixed_precision (PrecisionType | str):
            Mixed precision type e.g, fp16, bf16, etc. (disabled by default)
        use_cpu (bool):
            Whether to train using the CPU only even if CUDA is available.
        do_evaluate (bool):
            Whether to run evaluation when calling `Trainer.train`
        evaluate_with_generate (bool):
            Whether to use `generate()` in the evaluation step or not. (only applicable for generative models).
        metrics (List[str | MetricConfig]):
            A list of metrics. Depending on the `valid_metrics` in the specific MetricsHandler of the Trainer.
        metric_for_best_model (str):
            Reference metric key to watch for determining the best model. Recommended to have a {train. | evaluation.}
            prefix (e.g, evaluation.f1, train.accuracy, etc.) but if not, defaults to
            `evaluation.{metric_for_best_model}`.
        save_freq (int) (DEPRECATED):
            Deprecated and renamed to `save_steps`.
        save_enabled (bool):
            Whether to save checkpoints at all. `False` disables even the saves in-between the epochs.
        save_steps (int):
            Save the trainer outputs every `save_steps` steps. Leave as None to ignore saving in-between training steps.
            If set to a float value between 0 and 1, it will be interpreted as a fraction of the total steps.
        log_steps (int):
            Save training metrics every `log_steps` steps. If set to a float value between 0 and 1, it will be
            interpreted as a fraction of the total steps.
        checkpoints_dir (str):
            Path to the checkpoints' folder. The actual files will be saved under `{output_dir}/{checkpoints_dir}`.
        logs_dir (str):
            Path to the logs' folder. The actual log files will be saved under `{output_dir}/{logs_dir}`.
    """

    name: str = field(init=False, default="trainer")
    config_type: str = field(init=False, default=ConfigType.TRAINER)
    output_dir: str
    task: str | TaskType
    device: str = "cuda"
    num_epochs: int = None
    init_weights_from: str = None
    resume_from_checkpoint: bool | str | os.PathLike = None
    max_steps: int = None
    num_dataloader_workers: int = 0
    dataloader_shuffle: bool = True
    seed: int = 42
    optimizer: str | OptimizerType = None
    learning_rate: float = 2e-5
    weight_decay: float = 0.0
    lr_scheduler: str | LRSchedulerType = None
    lr_scheduler_kwargs: Dict[str, Any] = None
    lr_scheduling_steps: int = None
    batch_size: int = None
    eval_batch_size: int = None
    gradient_accumulation_steps: int = 1
    distributed: bool = False
    mixed_precision: PrecisionType | str | None = None
    use_cpu: bool = False
    do_evaluate: bool = True
    evaluate_with_generate: bool = True
    metrics: List[str | MetricConfig] = None
    metric_for_best_model: str = "loss"
    save_enabled: bool = True
    save_freq: int = "deprecated"
    save_steps: int = None
    log_steps: int = None
    checkpoints_dir: str = "checkpoints"
    logs_dir: str = "logs"

    def __post_init__(self):
        """
        Perform some argument sanitization and filtering here to avoid unexpected behavior in the trainer.
        The need for having this method is that some fields in the Trainer's config have correlations with each other
        and not controlling them can lead to conflicts.
        """
        super().__post_init__()
        # Validate `task`
        if self.task not in list(TaskType):
            raise ValueError(
                f"Invalid task `{self.task}` passed to `TrainerConfig`. "
                f"Available options are {TaskType.list()}",
            )
        # Validate `metric_for_best_model`
        if not (self.metric_for_best_model.startswith("evaluation") or self.metric_for_best_model.startswith("train")):
            if self.do_evaluate:
                self.metric_for_best_model = f"evaluation.{self.metric_for_best_model}"
            else:
                self.metric_for_best_model = f"train.{self.metric_for_best_model}"

        # Validate steps
        if isinstance(self.save_steps, int) and self.save_steps % self.gradient_accumulation_steps != 0:
            logger.warning(
                f"It's recommended to set a `save_steps` dividable by `gradient_accumulation_steps`, "
                f"otherwise, the saved model will have non-updated weights!\n"
                f"`save_steps={self.save_steps}`, `gradient_accumulation_steps={self.gradient_accumulation_steps}`"
            )

        # Validate deprecated fields
        if self.save_freq != "deprecated":
            logger.warning(
                "Trainer argument `save_freq` is deprecated! Use `save_steps` (number of training steps per save)."
                "Note that saving is also done at the end of each epoch unless you set `save_enabled` to `False` !"
            )

        # Distributed mode
        if self.distributed:
            logger.warning(
                "Distributed mode is experimental and might have bugs. "
                "Use with caution and don't forget to submit any issues!"
            )

        # Disable tokenizers parallelism if num_dataloader_workers is more than 1
        if self.num_dataloader_workers > 1:
            os.environ["TOKENIZERS_PARALLELISM"] = "false"



from dataclasses import dataclass

from ....configs import ModelConfig


@dataclass
class CraftTextDetectionConfig(ModelConfig):
    name = "craft_text_detection"
    text_threshold: float = 0.7
    link_threshold: float = 0.4
    low_text: float = 0.4
    poly: bool = False



from __future__ import annotations

import math
import os
import tempfile
from typing import Any, Callable, Dict

import pandas as pd
import torch
from huggingface_hub import create_repo, hf_hub_download, upload_file
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from tqdm.auto import tqdm

from ..configs import TrainerConfig
from ..constants import (
    DEFAULT_DATASET_CONFIG_FILE,
    DEFAULT_LR_SCHEDULER_FILE,
    DEFAULT_OPTIMIZER_FILE,
    DEFAULT_TRAINER_CONFIG_FILE,
    DEFAULT_TRAINER_CSV_LOG_FILE,
    DEFAULT_TRAINER_STATE_FILE,
    DEFAULT_TRAINER_SUBFOLDER,
    HEZAR_CACHE_DIR,
    TQDM_BAR_FORMAT,
    Backends,
    LRSchedulerType,
    OptimizerType,
    TaskType,
)
from ..data import Dataset, RangedSampler
from ..models import Model
from ..preprocessors import Preprocessor, PreprocessorsContainer
from ..utils import (
    Logger,
    colorize_text,
    dataloader_worker_init_fn,
    is_backend_available,
    sanitize_function_parameters,
    set_seed,
    verify_dependencies,
)
from .metrics_handlers import (
    Image2TextMetricHandler,
    MetricsHandler,
    SequenceLabelingMetricsHandler,
    SpeechRecognitionMetricsHandler,
    TextClassificationMetricsHandler,
    TextGenerationMetricsHandler,
)
from .trainer_utils import (
    AverageMeter,
    CSVLogger,
    TrainerState,
    get_distributed_logger,
    get_lr_scheduler_type,
    resolve_logdir,
    write_to_tensorboard,
)


if is_backend_available(Backends.ACCELERATE):
    from accelerate import Accelerator
else:
    raise ImportError("The package `accelerate` needs to be installed to use `Trainer`!")

logger = Logger(__name__)

optimizers = {
    OptimizerType.ADAM: torch.optim.Adam,
    OptimizerType.ADAMW: torch.optim.AdamW,
    OptimizerType.SDG: torch.optim.SGD,
}
lr_schedulers = {
    LRSchedulerType.LAMBDA: torch.optim.lr_scheduler.LambdaLR,
    LRSchedulerType.REDUCE_ON_PLATEAU: torch.optim.lr_scheduler.ReduceLROnPlateau,
    LRSchedulerType.STEP: torch.optim.lr_scheduler.StepLR,
    LRSchedulerType.MULTI_STEP: torch.optim.lr_scheduler.MultiStepLR,
    LRSchedulerType.ONE_CYCLE: torch.optim.lr_scheduler.OneCycleLR,
    LRSchedulerType.LINEAR: torch.optim.lr_scheduler.LinearLR,
    LRSchedulerType.EXPONENTIAL: torch.optim.lr_scheduler.ExponentialLR,
    LRSchedulerType.CYCLIC: torch.optim.lr_scheduler.CyclicLR,
    LRSchedulerType.SEQUENTIAL: torch.optim.lr_scheduler.SequentialLR,
    LRSchedulerType.POLYNOMIAL: torch.optim.lr_scheduler.PolynomialLR,
    LRSchedulerType.COSINE_ANEALING: torch.optim.lr_scheduler.CosineAnnealingLR,
}

task_to_metrics_handlers_mapping = {
    TaskType.TEXT_CLASSIFICATION: TextClassificationMetricsHandler,
    TaskType.SEQUENCE_LABELING: SequenceLabelingMetricsHandler,
    TaskType.IMAGE2TEXT: Image2TextMetricHandler,
    TaskType.SPEECH_RECOGNITION: SpeechRecognitionMetricsHandler,
    TaskType.TEXT_GENERATION: TextGenerationMetricsHandler,
}


class Trainer:
    """
    Base trainer class for training all Hezar models and all tasks. Usually you can use this class as-is, but for special
    cases you can also override any of the core methods in your own custom Trainer.

    Args:
        model (`Model` | `torch.nn.Module`): The main model to train and evaluate
        config (TrainerConfig): Training configuration and parameters
        train_dataset (Dataset): Train dataset
        eval_dataset (Dataset): Evaluation dataset
        data_collator: Collate function, usually included in the dataset object itself
        preprocessor: Preprocessor object(s)
        metrics_handler: Optional metrics handler
        optimizer (optim.Optimizer): Model optimizer
        lr_scheduler: Optional learning-rate scheduler
        accelerator (Accelerator) : Accelerator object for a customized distributed environment
    """

    trainer_subfolder = DEFAULT_TRAINER_SUBFOLDER
    trainer_config_file = DEFAULT_TRAINER_CONFIG_FILE
    trainer_csv_log_file = DEFAULT_TRAINER_CSV_LOG_FILE
    dataset_config_file = DEFAULT_DATASET_CONFIG_FILE
    trainer_state_file = DEFAULT_TRAINER_STATE_FILE
    optimizer_file = DEFAULT_OPTIMIZER_FILE
    lr_scheduler_file = DEFAULT_LR_SCHEDULER_FILE
    default_optimizer = OptimizerType.ADAM
    default_lr_scheduler = None
    _required_backends = [
        Backends.ACCELERATE,
        Backends.TENSORBOARD,
        Backends.DATASETS
    ]

    def __init__(
        self,
        model: Model,
        config: TrainerConfig,
        train_dataset: Dataset,
        eval_dataset: Dataset = None,
        data_collator: Callable = None,
        preprocessor: Preprocessor | PreprocessorsContainer = None,
        metrics_handler: MetricsHandler = None,
        optimizer: torch.optim.Optimizer = None,
        lr_scheduler=None,
        accelerator: "Accelerator" = None,
    ):
        # Check if all required dependencies are installed
        verify_dependencies(self._required_backends)

        # Configuration
        self.config = config
        self.device = "cuda" if torch.cuda.is_available() and not self.config.use_cpu else "cpu"

        # Setup hardware acceleration controller
        self.accelerator = accelerator or Accelerator(
            mixed_precision=self.config.mixed_precision,
            cpu=True if self.device == "cpu" else False,
            step_scheduler_with_optimizer=False,
            gradient_accumulation_steps=self.config.gradient_accumulation_steps,
        )

        # Setup logger
        self.logger = get_distributed_logger(__name__)

        # Configure datasets and data loaders
        self.train_dataset = train_dataset
        self.eval_dataset = eval_dataset
        self.data_collator = data_collator or getattr(self.train_dataset, "data_collator", None)

        # Configure steps
        self.num_batches = math.ceil(len(self.train_dataset) / self.config.batch_size)
        self.config.max_steps = (
            math.ceil(self.config.max_steps * self.num_batches) * self.config.num_epochs
            if isinstance(self.config.max_steps, float)
            and 0 < self.config.max_steps <= 1
            else self.config.max_steps
        )
        self.total_steps = min(
            self.config.max_steps or self.num_batches * self.config.num_epochs,
            self.num_batches * self.config.num_epochs
        )
        self.config.save_steps = (
            math.ceil(self.config.save_steps * self.total_steps)
            if isinstance(self.config.save_steps, float)
            and 0 < self.config.save_steps <= 1
            else self.config.save_steps
        )
        self.config.log_steps = (
            math.ceil(self.config.log_steps * self.total_steps)
            if isinstance(self.config.log_steps, float)
            and 0 < self.config.log_steps <= 1
            else self.config.log_steps
        )
        self.steps_in_epoch = min(self.num_batches, self.total_steps)
        self.config.save_steps = self.steps_in_epoch if not self.config.save_steps else self.config.save_steps
        self.saves_in_epoch = math.floor(self.steps_in_epoch / self.config.save_steps)

        # Setup checkpoint and state handler
        self.checkpoints_dir = os.path.join(self.config.output_dir, self.config.checkpoints_dir)
        self.state = self._create_trainer_state(self.config.resume_from_checkpoint)
        self.trainer_state_path = os.path.join(self.checkpoints_dir, self.trainer_state_file)
        self.logs_dir = self.state.logs_dir

        # Set determinism
        set_seed(self.config.seed)

        # Setup model and preprocessor(s)
        self.model = self._setup_model(model)
        if model.preprocessor is None:
            if preprocessor is not None:
                model.preprocessor = preprocessor
            else:
                raise ValueError(
                    "You must set a preprocessor for the model or pass the preprocessor parameter to the Trainer!"
                )

        # Setup optimizer and (optionally) lr scheduler
        self.optimizer, self.lr_scheduler = self._create_optimizers(optimizer, lr_scheduler)
        self.lr_scheduling_steps = self.config.lr_scheduling_steps or self.steps_in_epoch
        self.lr_scheduler_type = get_lr_scheduler_type(self.lr_scheduler, lr_schedulers)

        # Move objects to the accelerator
        self.model, self.optimizer, self.lr_scheduler = self.accelerator.prepare(
            self.model, self.optimizer, self.lr_scheduler
        )

        # Setup metrics handler and inner trackers for the trainer
        self.metrics_handler = metrics_handler or self._create_metrics_handler()
        self.train_loss_tracker = AverageMeter(
            name="train.loss",
            avg=self.state.loss_tracker_avg,
            sum=self.state.loss_tracker_sum,
            count=self.state.global_step,
        )

        # Setup logging properties
        self.tensorboard = SummaryWriter(log_dir=self.logs_dir)
        self.csv_logger = CSVLogger(logs_dir=self.logs_dir, csv_filename=self.trainer_csv_log_file)

    def _create_trainer_state(self, checkpoint: str = None):
        """
        Create Trainer's state attribute.

        Args:
            checkpoint: Optional checkpoint path or bool value to load the state from there

        Returns:
            A TrainerState object
        """
        trainer_state_path = os.path.join(self.checkpoints_dir, self.trainer_state_file)
        if checkpoint is not None and os.path.isfile(trainer_state_path):
            state = TrainerState.load(trainer_state_path)
            # Overwrite some fields if checkpoint is a path
            if os.path.isdir(checkpoint):
                step = os.path.basename(checkpoint)
                state.global_step = int(step) if step.isdigit() else self.state.global_step
                state.epoch = math.ceil(state.global_step / self.steps_in_epoch)
                state.epoch_step = state.global_step % self.steps_in_epoch

            if state.epoch_step == 0:
                state.epoch += 1
        else:
            state = TrainerState(
                epoch=1,
                total_epochs=self.config.num_epochs,
                metric_for_best_checkpoint=self.config.metric_for_best_model,
                logs_dir=resolve_logdir(os.path.join(self.config.output_dir, self.config.logs_dir)),
            )

        return state

    def _resolve_checkpoint_path(self, checkpoint: str | bool):
        if isinstance(checkpoint, bool) and checkpoint:
            checkpoint_name = str(self.state.global_step).zfill(len(str(self.total_steps)))
            checkpoint_path = os.path.join(self.checkpoints_dir, checkpoint_name)
        elif os.path.isdir(checkpoint):
            checkpoint_path = checkpoint
        else:
            raise ValueError(f"Checkpoint `{checkpoint}` is either invalid or does not exist!")

        return checkpoint_path

    def _setup_model(self, model: Model) -> Model:
        """
        Create and load the weights for the model. The weights will be loaded to the model depending
        on `config.resume_from_checkpoint` or `config.init_weights_from`.
        """
        if model is None:
            raise ValueError("`model` must be given to the Trainer!")

        # Maybe load from checkpoint
        if self.config.resume_from_checkpoint:
            checkpoint_path = self._resolve_checkpoint_path(self.config.resume_from_checkpoint)
            model_path = os.path.join(checkpoint_path, model.model_filename)
            if os.path.isdir(checkpoint_path) and os.path.isfile(model_path):
                model.load_state_dict(torch.load(model_path))
                self.logger.info(f"Resuming training from checkpoint at `{checkpoint_path}`")
        # Maybe load from pretrained weights locally or from a hub repo
        elif self.config.init_weights_from is not None and self.state.global_step == 0 and self.state.epoch == 1:
            if self.config.init_weights_from is not None:
                if os.path.isdir(self.config.init_weights_from):
                    model_path = os.path.join(self.config.init_weights_from, model.model_filename)
                else:
                    model_path = hf_hub_download(
                        self.config.init_weights_from,
                        filename=model.model_filename,
                        cache_dir=HEZAR_CACHE_DIR,
                    )
                model.load_state_dict(torch.load(model_path))

        return model

    def create_train_dataloader(self, dataset) -> DataLoader:
        """
        Create train data loader using a ranged sampler that can handle slicing data, shuffling, etc.
        """
        start_index = self.state.epoch_step * self.config.batch_size
        sampler = RangedSampler(
            dataset,
            self.config.batch_size,
            start_index=start_index,
            shuffle=self.config.dataloader_shuffle,
            seed=self.config.seed,
            drop_last=False,
        )
        worker_init_fn = dataloader_worker_init_fn(self.config.seed)

        train_dataloader = DataLoader(
            dataset=dataset,
            batch_size=self.config.batch_size,
            collate_fn=self.data_collator,
            sampler=sampler,
            num_workers=self.config.num_dataloader_workers,
            worker_init_fn=worker_init_fn,
        )

        train_dataloader = self.accelerator.prepare(train_dataloader)

        return train_dataloader

    def create_eval_dataloader(self, dataset) -> DataLoader:
        """
        Create eval data loader using a ranged sampler that can handle slicing data, shuffling, etc.
        """
        sampler = RangedSampler(
            dataset,
            self.config.eval_batch_size or self.config.batch_size,
            start_index=0,  # We don't support resumption for evaluation so we always start from zero
            shuffle=self.config.dataloader_shuffle,
            seed=self.config.seed,
            drop_last=False,
        )
        worker_init_fn = dataloader_worker_init_fn(self.config.seed)

        eval_dataloader = DataLoader(
            dataset=dataset,
            batch_size=self.config.eval_batch_size or self.config.batch_size,
            collate_fn=self.data_collator,
            sampler=sampler,
            num_workers=self.config.num_dataloader_workers,
            worker_init_fn=worker_init_fn,
        )

        eval_dataloader = self.accelerator.prepare(eval_dataloader)

        return eval_dataloader

    def _create_optimizers(self, optimizer: torch.optim.Optimizer = None, lr_scheduler=None):
        """
        Set up the optimizer and lr lr_scheduler if they're not already given

        Args:
            optimizer: If None do nothing and return it, otherwise build it using the train config
            lr_scheduler: If None do nothing and return it, otherwise build it using the train config

        Returns:
            Optimizer and lr_scheduler
        """
        if optimizer is None:
            optimizer_type = self.config.optimizer or self.default_optimizer
            optimizer = optimizers[optimizer_type](
                self.model.parameters(),
                lr=self.config.learning_rate,
                weight_decay=self.config.weight_decay,
            )

            if lr_scheduler is None:
                scheduler_name = self.config.lr_scheduler or self.default_lr_scheduler
                scheduler_kwargs = self.config.lr_scheduler_kwargs or {}
                if scheduler_name is None:
                    lr_scheduler = None
                else:
                    lr_scheduler = lr_schedulers[scheduler_name](optimizer, **scheduler_kwargs, verbose=True)

        if self.config.resume_from_checkpoint:
            checkpoint_path = self._resolve_checkpoint_path(self.config.resume_from_checkpoint)
            if os.path.isdir(checkpoint_path):
                optimizer_path = os.path.join(checkpoint_path, self.trainer_subfolder, self.optimizer_file)
                lr_scheduler_path = os.path.join(checkpoint_path, self.trainer_subfolder, self.lr_scheduler_file)
                if os.path.isfile(optimizer_path):
                    optimizer.load_state_dict(torch.load(optimizer_path))
                if lr_scheduler is not None and os.path.isfile(lr_scheduler_path):
                    lr_scheduler.load_state_dict(torch.load(lr_scheduler_path))

        return optimizer, lr_scheduler

    def _create_metrics_handler(self):
        """
        Setup MetricsHandler instance for the trainer

        Returns:
            A MetricsHandler subclass instance based on self.config.task
        """
        metrics_handler_cls = task_to_metrics_handlers_mapping[self.config.task]
        metrics_handler = metrics_handler_cls(metrics=self.config.metrics, trainer=self)  # noqa
        return metrics_handler

    def load_csv_logs(self, logs_dir=None):
        """
        Load the CSV log file
        Args:
            logs_dir: Path to logs directory, defaults to self.config.logs_dir

        Returns:
            Logs dictionary
        """
        logs_dir = logs_dir or self.logs_dir
        csv_path = os.path.join(logs_dir, self.trainer_csv_log_file)
        logs = pd.read_csv(csv_path)
        return logs.to_dict()

    def prepare_input_batch(self, input_batch) -> Dict[str, torch.Tensor]:
        """
        Every operation required to prepare the inputs for model forward like moving to device, permutations, etc.

        Args:
            input_batch: Raw input batch from the dataloader

        Returns:
            The proper input batch required by model forward
        """
        # Put inputs on device manually if accelerator is not available, otherwise it's taken care of by the accelerator
        if self.accelerator is None:
            input_batch = {k: v.to(self.device) if isinstance(v, torch.Tensor) else v for k, v in input_batch.items()}

        return input_batch

    def forward(self, input_batch):
        """
        Perform model forward on the input batch

        In special cases, one can override this method in their desired trainer.

        Args:
            input_batch: Input batch

        Returns:
            Model outputs
        """
        if isinstance(input_batch, torch.Tensor):
            outputs = self.model(input_batch)
        elif isinstance(input_batch, Dict):
            forward_inputs = sanitize_function_parameters(self.model.forward, input_batch)
            outputs = self.model(**forward_inputs)
        else:
            raise ValueError(
                f"`input_batch` must be a tensor or a dict-like object containing key/value pairs of tensors, "
                f"but got {type(input_batch)}"
            )
        if not isinstance(outputs, Dict):
            raise ValueError(f"Model outputs must be dict-like not `{type(outputs)}`")

        return outputs

    def compute_loss(self, model_outputs: Dict, labels: torch.Tensor, **kwargs) -> torch.Tensor:
        """
        Compute loss from model outputs

        Args:
            model_outputs: Logits from model outputs
            labels: Ground truth labels

        Returns:
            The loss tensor
        """
        compute_loss_inputs = sanitize_function_parameters(self.model.compute_loss, model_outputs, **kwargs)
        compute_loss_inputs["labels"] = labels

        loss = self.model.compute_loss(**compute_loss_inputs)

        return loss

    def optimization_step(self):
        """
        Perform optimization step
        """
        self.optimizer.step()
        self.optimizer.zero_grad()

    def training_step(self, input_batch: Dict[str, torch.Tensor]) -> Dict[str, Any]:
        """
        Train one batch of data and return loss and model outputs

        Args:
            input_batch: A batch of inputs to train

        Returns:
            Train step outputs including loss, logits, etc.
        """
        with self.accelerator.autocast():
            outputs = self.forward(input_batch)

        loss = self.compute_loss(outputs, **input_batch)

        self.accelerator.backward(loss)

        outputs["loss"] = loss.detach()

        return outputs

    def evaluation_step(self, input_batch: Dict[str, torch.Tensor]) -> Dict[str, Any]:
        """
        Evaluate one batch of data and return loss and model outputs

        Args:
            input_batch: A batch of inputs to evaluate

        Returns:
            Evaluation step outputs including loss, logits, etc.
        """
        with self.accelerator.autocast():
            outputs = self.forward(input_batch)

        loss = self.compute_loss(outputs, **input_batch)

        if self.model.is_generative and self.config.evaluate_with_generate:
            generate_inputs = sanitize_function_parameters(self.model.generate, input_batch)
            generated_ids = self.model.generate(**generate_inputs)
            outputs["logits"] = generated_ids["generated_ids"] if isinstance(generated_ids, dict) else generated_ids

        outputs["loss"] = loss

        return outputs

    def inner_training_loop(self, epoch_num: int):
        """
        Train the model for one epoch on the whole train dataset and verbose live metric values in the progress bar

        Args:
            epoch_num: Number of the current epoch

        Returns:
            Metrics averages through the full iteration
        """
        train_dataloader = self.create_train_dataloader(self.train_dataset)

        self.model.train()

        accumulated_loss = 0

        with tqdm(
            train_dataloader,
            initial=self.state.epoch_step,
            total=self.steps_in_epoch,
            unit="batch",
            desc=f"Epoch: {epoch_num}/{self.config.num_epochs} ",
            bar_format=TQDM_BAR_FORMAT,
            ascii=" #",
            disable=not self.accelerator.is_local_main_process,
        ) as iterator:
            for input_batch in iterator:
                # Handle early stopping
                if self.state.global_step >= self.total_steps:
                    break

                # Prepare inputs
                input_batch = self.prepare_input_batch(input_batch)

                # Training on one batch
                with self.accelerator.accumulate(self.model):
                    outputs = self.training_step(input_batch)
                    # Optimization step
                    self.optimization_step()

                # Update steps states
                self.state.global_step += 1
                self.state.epoch_step += 1

                # Gather outputs for metrics
                accumulated_loss += outputs["loss"].item()
                if (
                    self.state.epoch_step % self.config.gradient_accumulation_steps == 0
                    or self.state.epoch_step == self.steps_in_epoch
                ):
                    accumulated_loss = accumulated_loss / self.config.gradient_accumulation_steps
                    self.train_loss_tracker.update(accumulated_loss)
                    iterator.set_postfix(loss=self.train_loss_tracker.avg)

                    self.state.loss_tracker_avg = self.train_loss_tracker.avg
                    self.state.loss_tracker_sum = self.train_loss_tracker.sum
                    accumulated_loss = 0

                # Scheduler step
                if (
                    self.lr_scheduler is not None and
                    self.state.global_step % self.lr_scheduling_steps == 0 and
                    self.lr_scheduler_type != LRSchedulerType.REDUCE_ON_PLATEAU
                ):
                    self.lr_scheduler.step()

                # Save trainer outputs if `save_steps` is hit
                if self.config.save_steps and self.state.global_step % self.config.save_steps == 0:
                    ckpt_path_name = str(self.state.global_step).zfill(len(str(self.total_steps)))
                    self.save(os.path.join(self.checkpoints_dir, ckpt_path_name))
                    # Save Trainer state
                    self.state.save(self.trainer_state_path)

                # Log loss running mean
                if self.config.log_steps and self.state.global_step % self.config.log_steps == 0:
                    loss_mean = {"train.loss": self.train_loss_tracker.avg}
                    write_to_tensorboard(self.tensorboard, logs=loss_mean, step=self.state.global_step)

        return {"loss": self.train_loss_tracker.avg}

    def evaluate(self, eval_dataset: Dataset = None):
        """
        Evaluates the model on the whole eval dataset and verbose live metric values in the progress bar

        Args:
            eval_dataset: Any sized iterable like a Hezar Dataset, HuggingFace Dataset, Torch Dataset, etc.

        Returns:
            A dictionary of evaluation results computed by the metrics tracker
        """
        if eval_dataset is None and self.eval_dataset is None:
            raise ValueError(
                "Evaluation needs either passing the `eval_dataset` to the Trainer's `__init__` or `evaluate()`!"
            )
        eval_dataloader = self.create_eval_dataloader(eval_dataset or self.eval_dataset)

        self.metrics_handler.tracker.reset()
        self.model.eval()

        with tqdm(
            eval_dataloader,
            unit="batch",
            desc="Evaluating... ",
            bar_format=TQDM_BAR_FORMAT,
            ascii=" #",
            disable=not self.accelerator.is_local_main_process,
        ) as iterator:
            with torch.inference_mode():
                for step, input_batch in enumerate(iterator):
                    input_batch = self.prepare_input_batch(input_batch)
                    # Evaluation on one batch
                    outputs = self.evaluation_step(input_batch)
                    logits, labels = self.accelerator.gather_for_metrics((outputs["logits"], input_batch["labels"]))
                    # Compute metrics
                    evaluation_results = self.metrics_handler.compute_metrics(
                        logits.clone().detach().cpu(),
                        labels.clone().detach().cpu(),
                    )
                    evaluation_results["loss"] = self.accelerator.gather_for_metrics(outputs["loss"]).item()
                    # Gather outputs for metrics
                    self.metrics_handler.tracker.update(evaluation_results)
                    iterator.set_postfix(**self.metrics_handler.tracker.avg())

        return self.metrics_handler.tracker.avg()

    def print_info(self):
        """
        Print training info
        """

        def _print_info_line(key, value):
            line = f"  {colorize_text(key, 'bold')}: {colorize_text(str(value), 'italic')}"
            self.accelerator.print(line)

        header = f"{'*' * 20} Training Info {'*' * 20}"
        footer = "*" * len(header)
        info = {
            "Output Directory": self.config.output_dir,
            "Task": self.config.task,
            "Model": type(self.model).__name__,
            "Init Weights": self.config.init_weights_from or "N/A",
            "Device(s)": self.device,
            "Batch Size": self.config.batch_size,
            "Epochs": self.config.num_epochs,
            "Total Steps": self.total_steps,
            "Training Dataset": f"{self.train_dataset.__class__.__name__}({len(self.train_dataset)})",
            "Evaluation Dataset": f"{self.eval_dataset.__class__.__name__}({len(self.eval_dataset)})",
            "Optimizer": self.config.optimizer or self.default_optimizer,
            "Scheduler": self.config.lr_scheduler,
            "Initial Learning Rate": self.config.learning_rate,
            "Learning Rate Decay": self.config.weight_decay,
            "Number of Parameters": self.model.num_parameters,
            "Number of Trainable Parameters": self.model.num_trainable_parameters,
            "Mixed Precision": self.config.mixed_precision or "Full (fp32)",
            "Gradient Accumulation Steps": self.config.gradient_accumulation_steps,
            "Metrics": list(self.metrics_handler.metrics.keys()),
            "Save Steps": self.config.save_steps,
            "Log Steps": self.config.log_steps,
            "Checkpoints Path": self.checkpoints_dir,
            "Logs Path": self.logs_dir,
        }

        # Header
        self.accelerator.print(f"\n{colorize_text(header, 'bold')}\n")
        # Info
        [_print_info_line(k, v) for k, v in info.items()]
        # Footer
        self.accelerator.print(f"\n{colorize_text(footer, 'bold')}\n")

    def train(self, resume_from_checkpoint="deprecated"):
        """
        The full training process like training, evaluation, logging and saving model checkpoints.

        The steps are as follows:
            - The following is run for `self.config.num_epochs` times
                - Run the training loop on the train dataset
                - Save checkpoints
                - Run evaluation on the evaluation dataset
                - Apply LR scheduling of a LR Scheduler is available
                - Gather all metrics outputs
                - Save the trainer state
                - Write logs to tensorboard, csv, etc.
        """
        if resume_from_checkpoint != "deprecated":
            raise ValueError(
                "Setting `resume_from_checkpoint` in `Trainer.train(resume_from_checkpoint=...)` is deprecated. "
                "You have to set it in the trainer's config!"
            )

        self.print_info()

        for epoch in range(self.state.epoch, self.config.num_epochs + 1):
            self.accelerator.print()
            self.state.epoch = epoch
            metrics_logs = {}

            # Train on the whole training set
            training_results = self.inner_training_loop(epoch)
            metrics_logs.update({"train.loss": training_results["loss"]})

            # Save checkpoint
            if self.accelerator.is_local_main_process and self.config.save_enabled:
                ckpt_path_name = str(self.state.global_step).zfill(len(str(self.total_steps)))
                self.save(os.path.join(self.checkpoints_dir, ckpt_path_name))

            if self.config.do_evaluate:
                # Evaluate the model on the evaluation set
                evaluation_results = self.evaluate(self.eval_dataset)
                evaluation_logs = {
                    f"evaluation.{metric_name}": value for metric_name, value in evaluation_results.items()
                }
                metrics_logs.update(evaluation_logs)

            # LR scheduler step (only for reduce on plateau)
            if self.lr_scheduler_type == LRSchedulerType.REDUCE_ON_PLATEAU:
                self.lr_scheduler.step(metrics_logs[self.config.metric_for_best_model])

            # Update trainer state
            self.state.epoch = epoch
            self.state.epoch_step = 0
            self.state.update_best_results(
                metric_value=metrics_logs[self.config.metric_for_best_model],
                objective=self.metrics_handler.objective,
                step=self.state.global_step,
            )

            # Log everything
            self.log(metrics_logs, self.state.global_step)

            # Early stopping
            if self.state.global_step == self.total_steps:
                break

        self.logger.info("Training done!")

    def log(self, logs: Dict[str, Any], step: int):
        """
        Log metrics results
        """
        # Log to tensorboard
        write_to_tensorboard(self.tensorboard, logs, step)

        # Log to CSV
        self.csv_logger.write(logs, step)

        # Save trainer state
        self.state.save(
            os.path.join(
                self.checkpoints_dir,
                self.trainer_state_file,
            )
        )

    def save(
        self,
        path: str,
        config_filename: str = None,
        model_filename: str = None,
        model_config_filename: str = None,
        subfolder: str = None,
        dataset_config_file: str = None,
        optimizer_file: str = None,
        lr_scheduler_file: str = None,
    ):
        """
        Save the trainer and relevant files to a path.

        Files to save are train config, model weights, model config, preprocessor files and preprocessor config.

        Args:
            path: A directory to save everything
            config_filename: Config file name
            model_filename: Model file name
            model_config_filename: Model config file name
            subfolder: Optional sub-folder
            dataset_config_file: Dataset config file name
            optimizer_file: Optimizer state file name
            lr_scheduler_file: LR scheduler file name
        """
        config_filename = config_filename or self.trainer_config_file
        subfolder = subfolder or self.trainer_subfolder
        dataset_config_file = dataset_config_file or self.dataset_config_file
        optimizer_file = optimizer_file or self.optimizer_file
        lr_scheduler_file = lr_scheduler_file or self.lr_scheduler_file

        self.config.save(path, filename=config_filename, subfolder=subfolder)
        self.model.save(path, filename=model_filename, config_filename=model_config_filename)

        torch.save(self.optimizer.state_dict(), os.path.join(path, subfolder, optimizer_file))

        if self.lr_scheduler is not None:
            torch.save(self.lr_scheduler.state_dict(), os.path.join(path, subfolder, lr_scheduler_file))

        if isinstance(self.train_dataset, Dataset):
            self.train_dataset.config.save(path, filename=dataset_config_file, subfolder=subfolder)

    def push_to_hub(
        self,
        repo_id: str,
        config_filename: str = None,
        push_model: bool = True,
        push_optimizer: bool = True,
        push_logs: bool = True,
        model_filename: str = None,
        model_config_filename: str = None,
        optimizer_filename: str = None,
        subfolder: str = None,
        dataset_config_filename: str = None,
        commit_message: str = None,
        private: bool = False,
    ):
        """
        Push everything to the Hub

        Args:
            repo_id: Path to hub
            config_filename: Trainer config file name
            push_model: Whether to push the model
            push_optimizer: Whether to push the optimizer
            push_logs: Whether to push training logs
            model_filename: Model file name
            optimizer_filename: Optimizer file name
            model_config_filename: Model config file name
            subfolder: Path to Trainer files
            dataset_config_filename: Dataset config file name
            commit_message: Commit message for the push
            private: Whether to create a private repo if it doesn't exist already
        """
        config_filename = config_filename or self.trainer_config_file
        subfolder = subfolder or self.trainer_subfolder
        dataset_config_file = dataset_config_filename or self.dataset_config_file
        optimizer_filename = optimizer_filename or self.optimizer_file

        # create remote repo
        create_repo(repo_id, repo_type="model", exist_ok=True, private=private)

        if not commit_message:
            commit_message = "Hezar: Upload training files"

        # upload train files
        self.config.push_to_hub(
            repo_id,
            filename=config_filename,
            subfolder=subfolder,
            private=private,
            commit_message=commit_message,
        )

        # Upload train dataset config
        if isinstance(self.train_dataset, Dataset):
            self.train_dataset.config.push_to_hub(
                repo_id,
                filename=dataset_config_file,
                subfolder=subfolder,
                private=private,
                commit_message=commit_message
            )

        # Upload model
        if push_model:
            self.model.push_to_hub(
                repo_id,
                filename=model_filename,
                config_filename=model_config_filename,
                commit_message=commit_message,
                private=private,
            )

        # Upload optimizer state
        if push_optimizer:
            optimizer_path = os.path.join(tempfile.mkdtemp(), optimizer_filename)
            torch.save(self.optimizer.state_dict(), optimizer_path)
            upload_file(
                path_or_fileobj=optimizer_path,
                path_in_repo=os.path.join(self.trainer_subfolder, optimizer_filename),
                repo_id=repo_id,
                commit_message=commit_message,
            )
        # Upload logs
        if push_logs:
            upload_file(
                path_or_fileobj=self.csv_logger.save_path,
                path_in_repo=os.path.join(self.trainer_subfolder, self.trainer_csv_log_file),
                repo_id=repo_id,
                commit_message=commit_message,
            )



import os

from huggingface_hub import HfApi, Repository

from ..constants import HEZAR_CACHE_DIR, RepoType
from ..utils.logging import Logger


__all__ = [
    "get_local_cache_path",
    "exists_in_cache",
    "exists_on_hub",
    "clone_repo",
    "list_repo_files",
    "get_state_dict_from_hub",
    "clean_cache",
]

logger = Logger(__name__)


def get_local_cache_path(repo_id, repo_type):
    """
    Given the hub path and repo type, configure the local path to save everything e.g, ~/.hezar/models/<repo_name>

    Args:
        repo_id: Repo name or id
        repo_type: Repo type e.g, model, dataset, etc

    Returns:
        Path to local cache directory
    """
    repo_owner, repo_name = repo_id.split("/")
    cache_path = f"{HEZAR_CACHE_DIR}/{repo_type}s--{repo_owner}--{repo_name}"
    return cache_path


def exists_in_cache(hub_path, repo_type="model"):
    cache_path = get_local_cache_path(hub_path, repo_type)
    return os.path.exists(cache_path)


def exists_on_hub(hub_path: str, repo_type="model"):
    """
    Determine whether the repo exists on the hub or not

    Args:
        hub_path: Repo name or id
        repo_type: Repo type like model, dataset, etc.

    Returns:
        True or False
    """
    author, repo_name = hub_path.split("/")
    api = HfApi()
    if repo_type == "model":
        paths = list(iter(api.list_models(author=author)))
    elif repo_type == "dataset":
        paths = list(iter(api.list_datasets(author=author)))
    elif repo_type == "space":
        paths = list(iter(api.list_spaces(author=author)))
    else:
        raise ValueError(f"Unknown type: {repo_type}! Use `model`, `dataset`, `space`, etc.")

    return hub_path in [path.id for path in paths]


def clone_repo(repo_id: str, save_path: str, **kwargs):
    """
    Clone a repo on the hub to local directory

    Args:
        repo_id: Repo name or id
        save_path: Path to clone the repo to

    Returns:
        the local path to the repo
    """
    repo = Repository(local_dir=save_path, clone_from=repo_id, **kwargs)
    return repo.local_dir


def list_repo_files(hub_or_local_path: str, subfolder: str = None, repo_type: str | RepoType = RepoType.MODEL):
    """
    List all files in a Hub or local model repo

    Args:
        hub_or_local_path: Path to hub or local repo
        subfolder: Optional subfolder path
        repo_type: Repo type of either dataset or model

    Returns:
        A list of all file names
    """
    if os.path.isdir(hub_or_local_path):
        files_itr = os.walk(hub_or_local_path)
        files = []
        for root, dirs, files_ in files_itr:
            for file in files_:
                files.append(os.path.relpath(os.path.join(root, file), hub_or_local_path))
    else:
        files = HfApi().list_repo_files(hub_or_local_path, repo_type=str(repo_type))

    if subfolder:
        files = [os.path.relpath(f, subfolder) for f in files if subfolder in f]
    return files


def get_state_dict_from_hub(hub_id, filename, subfolder=None):
    """
    Load a state dict from a repo on the HF Hub. Works on any repo no matter the library.

    Args:
        hub_id: Path to repo id
        filename: Weights file name
        subfolder: Optional subfolder in the repo

    Returns:
        A PyTorch state dict obj
    """
    import torch

    api = HfApi()

    subfolder = subfolder or ""

    # Download or load the cached file
    weights_file = api.hf_hub_download(
        repo_id=hub_id,
        filename=filename,
        subfolder=subfolder,
        cache_dir=HEZAR_CACHE_DIR,
    )

    state_dict = torch.load(weights_file)

    return state_dict


def clean_cache(cache_dir: str = None, delay: int = 10):
    """
    Clean the whole cache directory of Hezar

    Args:
        cache_dir: Optionally provide the cache dir path or the default cache dir will be used otherwise.
        delay: How many seconds to wait before performing the deletion action
    """
    import shutil
    import time

    cache_dir = cache_dir or HEZAR_CACHE_DIR

    logger.warning(f"Attempting to delete the files in `{cache_dir}` in {delay} seconds...")
    time.sleep(delay)

    shutil.rmtree(cache_dir)
    logger.info(f"Successfully deleted `{cache_dir}`!")



r"""
Hezar uses a registry system in a way that for any core module like model, dataset, etc. there is an entry in its
specific registry. These registries are simple python dictionaries that map a module's name to its class and its config
class. These registries are initialized here and filled automatically when you import hezar or a registry itself.

Examples:
    >>> # read models registry
    >>> from hezar.registry import models_registry
    >>> print(models_registry)
    {'distilbert_mask_filling': {'module_class': <class 'hezar.models.mask_filling.distilbert.distilbert_mask_filling.DistilBertMaskFilling'>,
    'config_class': <class 'hezar.models.mask_filling.distilbert.distilbert_mask_filling_config.DistilBertMaskFillingConfig'>},
    'description': 'Optional model description here...'}

    >>> # add a model class to models_registry
    >>> from hezar.models import Model, register_model
    >>> @register_model(name="my_awesome_model", config_class=MyAwesomeModelConfig, description="My Awesome Model!")
    >>> class MyAwesomeModel(Model):
    ...    def __init__(config: MyAwesomeModelConfig):
    ...        ...

Keep in mind that registries usually don't need to be used directly. There is a bunch of functions to build modules
using a module's registry name in `hezar.builders` module. See the file `builders.py` for more info.

Note: In case of adding a new registry container, make sure to add to `__all__` below!
"""

from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, Optional, Type


if TYPE_CHECKING:
    from .configs import (
        Config,
        DatasetConfig,
        EmbeddingConfig,
        MetricConfig,
        ModelConfig,
        PreprocessorConfig,
    )

from .utils import Logger


__all__ = [
    "register_model",
    "register_preprocessor",
    "register_dataset",
    "register_embedding",
    "register_metric",
    "Registry",
    "models_registry",
    "preprocessors_registry",
    "datasets_registry",
    "embeddings_registry",
    "metrics_registry",
]

logger = Logger(__name__)


@dataclass
class Registry:
    module_class: type
    config_class: type = None
    description: Optional[str] = None


models_registry: Dict[str, Registry] = {}
preprocessors_registry: Dict[str, Registry] = {}
datasets_registry: Dict[str, Registry] = {}
embeddings_registry: Dict[str, Registry] = {}
metrics_registry: Dict[str, Registry] = {}


def _register_module(
    cls: Type,
    registry: Dict[str, Registry],
    module_name: str,
    config_class: Type["Config"],
    description: str = None
):
    """
    Add module to the registry.

    Args:
        cls: The module class
        registry: Module's registry container
        module_name: Module's registry name (key)
        config_class: Module's config class
        description: Optional description for the module
    """
    if module_name in registry:
        logger.warning(f"`{module_name}` is already registered. Overwriting...")

    if config_class.name != module_name:
        raise ValueError(
            f"Module's registry name and `config.name` are not compatible for `{cls.__name__}`\n"
            f"Registry name: {module_name}\n"
            f"{config_class.__name__}.name: {config_class.name}"
        )
    registry[module_name] = Registry(module_class=cls, config_class=config_class, description=description)


def register_model(model_name: str, config_class: Type["ModelConfig"], description: str = None):
    """
    A class decorator that adds the model class and the config class to the `models_registry`

    Args:
        model_name: Model's registry name e.g, `bert_sequence_labeling`
        config_class: Model's config class e.g, `BertSequenceLabelingConfig`. This parameter must be the config class
            itself not a config instance!
        description: Optional model description
    """

    def register(cls):
        _register_module(cls, models_registry, model_name, config_class, description)
        return cls

    return register


def register_dataset(dataset_name: str, config_class: Type["DatasetConfig"], description: str = None):
    """
    A class decorator that adds the dataset class and the config class to the `datasets_registry`

    Args:
        dataset_name: Dataset's registry name e.g, `text_classification`.
        config_class: Dataset's config class e.g, `TextClassificationDatasetConfig`. This parameter must be the config
            class itself not a config instance!
        description: Optional dataset description
    """

    def register(cls):
        _register_module(cls, datasets_registry, dataset_name, config_class, description)
        return cls

    return register


def register_preprocessor(preprocessor_name: str, config_class: Type["PreprocessorConfig"], description: str = None):
    """
    A class decorator that adds the preprocessor class and the config class to the `preprocessors_registry`

    Args:
        preprocessor_name: Preprocessor's registry name e.g, `bpe_tokenizer`.
        config_class: Preprocessor's config class e.g, BPEConfig. This parameter must be the config
            class itself not a config instance!
        description: Optional preprocessor description
    """

    def register(cls):
        _register_module(cls, preprocessors_registry, preprocessor_name, config_class, description)
        return cls

    return register


def register_embedding(embedding_name: str, config_class: Type["EmbeddingConfig"], description: str = None):
    """
    A class decorator that adds the embedding class and the config class to the `embeddings_registry`

    Args:
        embedding_name: Embedding's registry name e.g, `word2vec_cbow`.
        config_class: Embedding's config class e.g, Word2VecCBOWConfig. This parameter must be the config
            class itself not a config instance!
        description: Optional embedding description
    """

    def register(cls):
        _register_module(cls, embeddings_registry, embedding_name, config_class, description)
        return cls

    return register


def register_metric(metric_name: str, config_class: Type["MetricConfig"], description: str = None):
    """
    A class decorator that adds the metric class and the config class to the `metrics_registry`

    Args:
        metric_name: Metric registry name e.g, `f1`
        config_class: Metric config class
        description: Optional metric description
    """

    def register(cls):
        _register_module(cls, metrics_registry, metric_name, config_class, description)
        return cls

    return register



from collections import namedtuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import models

from ....constants import Backends
from ....registry import register_model
from ...model import Model
from ...model_outputs import TextDetectionOutput
from .craft_text_detection_config import CraftTextDetectionConfig
from .craft_utils import adjust_result_coordinates, get_detection_boxes, polys2boxes


_required_backends = [
    Backends.OPENCV,
    Backends.PILLOW,
    Backends.TORCHVISION,
]


def init_weights(modules):
    for m in modules:
        if isinstance(m, nn.Conv2d):
            nn.init.xavier_uniform_(m.weight.data)
            if m.bias is not None:
                m.bias.data.zero_()
        elif isinstance(m, nn.BatchNorm2d):
            m.weight.data.fill_(1)
            m.bias.data.zero_()
        elif isinstance(m, nn.Linear):
            m.weight.data.normal_(0, 0.01)
            m.bias.data.zero_()


@register_model("craft_text_detection", config_class=CraftTextDetectionConfig)
class CraftTextDetection(Model):
    """
    CRAFT for text detection. Copied from the original implementation at https://github.com/clovaai/CRAFT-pytorch
    """
    required_backends = _required_backends

    def __init__(self, config: CraftTextDetectionConfig, **kwargs):
        super(CraftTextDetection, self).__init__(config=config, **kwargs)

        """ Base network """
        self.basenet = VGG16BN()

        """ U network """
        self.upconv1 = DoubleConv(1024, 512, 256)
        self.upconv2 = DoubleConv(512, 256, 128)
        self.upconv3 = DoubleConv(256, 128, 64)
        self.upconv4 = DoubleConv(128, 64, 32)

        num_class = 2
        self.conv_cls = nn.Sequential(
            nn.Conv2d(32, 32, kernel_size=3, padding=1), nn.ReLU(inplace=True),
            nn.Conv2d(32, 32, kernel_size=3, padding=1), nn.ReLU(inplace=True),
            nn.Conv2d(32, 16, kernel_size=3, padding=1), nn.ReLU(inplace=True),
            nn.Conv2d(16, 16, kernel_size=1), nn.ReLU(inplace=True),
            nn.Conv2d(16, num_class, kernel_size=1),
        )

        init_weights(self.upconv1.modules())
        init_weights(self.upconv2.modules())
        init_weights(self.upconv3.modules())
        init_weights(self.upconv4.modules())
        init_weights(self.conv_cls.modules())

    def forward(self, pixel_values, ratio_values=None):
        """ Base network """
        sources = self.basenet(pixel_values)

        """ U network """
        y = torch.cat([sources[0], sources[1]], dim=1)
        y = self.upconv1(y)

        y = F.interpolate(y, size=sources[2].size()[2:], mode='bilinear', align_corners=False)
        y = torch.cat([y, sources[2]], dim=1)
        y = self.upconv2(y)

        y = F.interpolate(y, size=sources[3].size()[2:], mode='bilinear', align_corners=False)
        y = torch.cat([y, sources[3]], dim=1)
        y = self.upconv3(y)

        y = F.interpolate(y, size=sources[4].size()[2:], mode='bilinear', align_corners=False)
        y = torch.cat([y, sources[4]], dim=1)
        feature = self.upconv4(y)

        y = self.conv_cls(feature).permute(0, 2, 3, 1)

        return {"logits": y, "ratio_values": ratio_values}

    def preprocess(self, inputs, **kwargs):
        processed_outputs = self.preprocessor.image_processor(inputs, **kwargs)
        return processed_outputs

    def post_process(
        self,
        model_outputs: dict,
        text_threshold: float = None,
        link_threshold: float = None,
        low_text: float = None,
        poly: bool = False,
    ):
        text_threshold = text_threshold or self.config.text_threshold
        link_threshold = link_threshold or self.config.link_threshold
        low_text = low_text or self.config.low_text

        logits = model_outputs["logits"]
        ratio_values = model_outputs["ratio_values"]

        results = []

        for output, ratio in zip(logits, ratio_values):
            # make score and link map
            score_text = output[:, :, 0].cpu().data.numpy()
            score_link = output[:, :, 1].cpu().data.numpy()

            # Post-processing
            boxes, polys, mapper = get_detection_boxes(
                score_text,
                score_link,
                text_threshold,
                link_threshold,
                low_text,
                poly,
            )

            # coordinate adjustment
            boxes = adjust_result_coordinates(boxes, ratio, ratio)
            for k in range(len(polys)):
                if polys[k] is None:
                    polys[k] = boxes[k]
            boxes = polys2boxes(polys)
            results.append(TextDetectionOutput(boxes=boxes))
        return results


class VGG16BN(nn.Module):
    def __init__(self):
        super(VGG16BN, self).__init__()
        vgg_pretrained_features = models.vgg16_bn().features

        self.slice1 = nn.Sequential()
        self.slice2 = nn.Sequential()
        self.slice3 = nn.Sequential()
        self.slice4 = nn.Sequential()
        self.slice5 = nn.Sequential()
        for x in range(12):  # conv2_2
            self.slice1.add_module(str(x), vgg_pretrained_features[x])
        for x in range(12, 19):  # conv3_3
            self.slice2.add_module(str(x), vgg_pretrained_features[x])
        for x in range(19, 29):  # conv4_3
            self.slice3.add_module(str(x), vgg_pretrained_features[x])
        for x in range(29, 39):  # conv5_3
            self.slice4.add_module(str(x), vgg_pretrained_features[x])

        # fc6, fc7 without atrous conv
        self.slice5 = nn.Sequential(
            nn.MaxPool2d(kernel_size=3, stride=1, padding=1),
            nn.Conv2d(512, 1024, kernel_size=3, padding=6, dilation=6),
            nn.Conv2d(1024, 1024, kernel_size=1)
        )

        init_weights(self.slice1.modules())
        init_weights(self.slice2.modules())
        init_weights(self.slice3.modules())
        init_weights(self.slice4.modules())
        init_weights(self.slice5.modules())  # no pretrained model for fc6 and fc7

    def forward(self, x):
        h = self.slice1(x)
        h_relu2_2 = h
        h = self.slice2(h)
        h_relu3_2 = h
        h = self.slice3(h)
        h_relu4_3 = h
        h = self.slice4(h)
        h_relu5_3 = h
        h = self.slice5(h)
        h_fc7 = h
        vgg_outputs = namedtuple("VggOutputs", ['fc7', 'relu5_3', 'relu4_3', 'relu3_2', 'relu2_2'])
        out = vgg_outputs(h_fc7, h_relu5_3, h_relu4_3, h_relu3_2, h_relu2_2)
        return out


class DoubleConv(nn.Module):
    def __init__(self, in_ch, mid_ch, out_ch):
        super(DoubleConv, self).__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(in_ch + mid_ch, mid_ch, kernel_size=1),
            nn.BatchNorm2d(mid_ch),
            nn.ReLU(inplace=True),
            nn.Conv2d(mid_ch, out_ch, kernel_size=3, padding=1),
            nn.BatchNorm2d(out_ch),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        x = self.conv(x)
        return x
